(** Syntax *)

open Library

type svalue =
    Number of scmnum
  | Boolean of bool
  | Character of char
  | String of string
  | Symbol of string
  | Null
  | Undef

and exp =
    Val of svalue
  | Var of  string
  | Lambda of (string list * string option) * exp
  | App of exp * exp list
  | Let of bind list * exp
  | Letrec of bind list * exp
  | If  of exp * exp * exp
  | Set of string * exp
and bind = { id : string ; body : exp }

let count = ref 0
let new_name =
  function _ -> 
    begin 
      count := !count + 1;
      "v" ^ string_of_int !count
    end


let debug str = failwith ("syntax.ml : " ^ str)


let trim_letrec e = 
  match e with
    Letrec([],e2) -> e2
  | Let([],e2) -> e2
  | _ -> e

let rec program2exp (prog:PreSyntax.program) : exp =
  let rec is_coms cs =
    match cs with
      `Command e::tl ->
	is_coms tl
    | [] -> true
    | _ -> false
  in 
  let rec rem_begin dclist = 
    List.fold_right (@)
      (List.map 
	 (function `Command e -> [`Command e]
	   | `Define d -> [`Define d]
	   | `P_Begin lst -> rem_begin lst)
	 dclist)
      []
  in 
  let rec get_defs ds =
    match ds with
      `Command e::tl ->
	if is_coms ds
	then
	  ([],List.map (function `Command e -> e | _ -> failwith "parser") ds)
	else 
	  failwith "parser"
    | `Define d::tl -> 
	let (ds2,cs) = get_defs tl in
	(d::ds2,cs)
    | [] ->
	([],[])
    | _ -> failwith "parser"
  in 
  (* ȥޥʬ롣 r5rsͤϤʤɡȤꤢ  *)
  let (defs, coms) = get_defs (rem_begin prog) in
  
  trim_letrec (Letrec(
  begin
    List.map 
      (function 
	  `DefineV(v,expression) -> 
	    {id=v; body=expression2exp expression}
	| `DefineF(v,(args,tail),body) -> 
	    {id=v; body=Lambda((args,tail),body2exp body)}
	| _ -> failwith "not implemented : syntaxl.ml") defs
  end,
  begin2let (List.map expression2exp coms))
 )

and body2exp (b:PreSyntax.body) : exp =
  match b with
    (dlst,elst) ->
      trim_letrec (Letrec(
      begin
	List.map 
	  (function 
	      `DefineV(v,expression) -> 
		{id=v; body=expression2exp expression}
	    | `DefineF(v,(args,tail),body) -> 
		{id=v; body=Lambda((args,tail),body2exp body)}
	    | _ -> failwith "not implemented : syntaxl.ml") dlst
      end,
      begin2let (List.map expression2exp elst)))
	
and expression2exp (e:PreSyntax.expression) :exp =
  match e with
    `Variable v -> Var v
  | `Literal l -> literal2exp l
  | `Procedure_call (e,args) -> App(expression2exp e,
				    List.map expression2exp args)
  | `Lambda_expression((args,tail),body) -> Lambda((args,tail),body2exp body)
  | `Conditional_if(e1,e2,None) -> If(expression2exp e1,
				      expression2exp e2,
				      Val Undef)
  | `Conditional_if(e1,e2,Some e3) -> If(expression2exp e1,
				    expression2exp e2,
				    expression2exp e3)
  | `Assignment(v,e) -> Set(v,expression2exp e)
  | `Derived_expression dexp -> d_exp2exp dexp
  | `Macro_use -> debug "macro_use"
  | `Macro_block -> debug "macro_block"
  | `Do_Nothing -> Val Undef
	
and d_exp2exp (dexp:PreSyntax.derived_expression) :exp =
  match dexp with
    `Cond(cond_clause,elses) ->
      begin 
	let rec iter ccl =
	  match ccl with
	    (test,exps)::tl ->
	      If(expression2exp test,
		 begin2let (List.map expression2exp exps),
		 iter tl)
	  | [] ->
	      begin2let (List.map expression2exp elses)
	in
	iter cond_clause
      end
  | `Case _ -> debug "case"
  | `And exps -> 
      begin 
	let rec iter exps =
	  match exps with
	    [] -> Val (Boolean true)
	  | [hd] -> 
	      let v = new_name () in
	      Let([{id=v; body=expression2exp hd}],
		  If(Var v,
		     Var v,
		     Val(Boolean false)))
	  | hd::tl ->
	      let v = new_name () in
	      Let([{id=v; body=expression2exp hd}],
		  If(Var v,
		     iter tl,
		     Val(Boolean false)))
	in iter exps
      end
  | `Or exps ->
      begin 
	let rec iter exps =
	  match exps with
	    [] -> Val (Boolean false)
	  | hd::tl ->
	      let v = new_name () in
	      Let([{id=v; body=expression2exp hd}],
		  If(Var v,
		     Var v,
		     iter tl))
	in iter exps
      end
  | `Let(blst,body) ->
      Let(List.map (function (v,body) -> {id=v; body=expression2exp body}) blst,
	  body2exp body)
  | `NamedLet _ ->  debug "named let"
  | `LetStar(blst,body) ->
      begin
	let rec iter blst =
	  match blst with
	    (v,b)::tl -> Let([{id=v; body=expression2exp b}],
				iter tl)
	  | [] -> body2exp body
	in 
	iter blst
      end
  | `Letrec(blst,body) ->
      Letrec(List.map (function (v,body) -> {id=v; body=expression2exp body}) blst,
	     body2exp body)
  | `Begin exps ->  begin2let (List.map expression2exp exps)
  | `Do(iters,(test,exp),command) ->
      let fname = new_name () in
      Let(List.map (function (v,i,e) -> {id=v; body=expression2exp i}) iters,
	  Letrec([{id=fname; 
		   body= Lambda((List.map (function (v,_,_) -> v) iters,None),
				If(expression2exp test, begin2let (List.map expression2exp exp), 
				   begin2let ((List.map expression2exp command)
					      @[App(Var fname,List.map (function (_,_,e) -> expression2exp e) iters)])))}],
		 App(Var fname,List.map (function (v,_,_) -> Var v) iters)))
  | `Delay -> debug "delay"
  | `Quasiquotation -> debug "quasiquotation"
	  
and begin2let (exps:exp list) : exp =
  match exps with
    [] -> Val Undef
  | [hd] -> hd
  | hd::tl -> 
      let v = new_name () in 
      Let([{id=v; body=hd}],
	  begin2let tl)
     
      
and literal2exp (l:PreSyntax.literal) : exp =
  match l with
    `Datum d -> datum2exp d
  | `Self_evaluating s -> self2exp s

and datum2exp d : exp=
  match d with
    `Boolean b -> Val (Boolean b)
  | `Number  n -> Val (Number n)
  | `Character c -> Val (Character c)
  | `String s -> Val (String s)
  | `Symbol s -> Val (Symbol s)
  | `List l   -> 
      begin
      let rec iter l =
	match l with
	  hd::tl -> App(Var("cons"),[datum2exp hd;iter tl])
	| [] -> Val Null
      in iter l
      end
  | `Vector l -> debug "not implemented abbrev"
  | `Abbrev ab -> debug "not implemented abbrev"
	
and self2exp (self:PreSyntax.self_evaluating) :exp =
  match self with
    `Boolean b -> Val (Boolean b)
  | `Number n -> Val (Number n )
  | `Character c -> Val (Character c)
  | `String s -> Val (String s)


(*-----------------------------------------------------------------------------*)
(*-----------------------------------------------------------------------------*)
(*-----------------------------------------------------------------------------*)
(*-----------------------------------------------------------------------------*)


(* alpha Ѵ *)						       
let rec alpha_conv vpair = 
  function 
      Val(sval) -> Val(alpha_conv_sval vpair sval)
    | (Var(name) as v) -> 
	begin 
	  try 
	    Var(List.assoc name vpair)
	  with _ -> 
	    failwith ("Warning : unknown variable " ^ name ^"\n")
	end
    | App(f,args) -> App(alpha_conv vpair f,
			 List.map (alpha_conv vpair) args)
    | Lambda((args,Some tail),e) -> 
	let new_vars = List.map new_name args in
	let new_tail = new_name tail in
	Lambda((new_vars,Some new_tail),
	       alpha_conv 
		 (List.append 
		    ((tail,new_tail)::(List.combine args new_vars)) vpair) e
	      )
    | Lambda((args,None),e) -> 
	let new_vars = List.map new_name args in
	Lambda((new_vars,None),
	       alpha_conv 
		 (List.append 
		    (List.combine args new_vars) vpair) e
	      )
    | Let(blst,exp) ->
	let for_bind bnd =
	  {id = new_name bnd.id; body = (alpha_conv vpair bnd.body) } in 
	let newblst = List.map for_bind blst in
	let get_vars bnd = bnd.id in
	Let(newblst,
	    alpha_conv
	      (List.append 
		 (List.combine (List.map get_vars blst) (List.map get_vars newblst))
		 vpair) exp)
    | Letrec(blst,exp) ->
	let new_vpair = List.map (function x -> (x.id, new_name ())) blst in
	let newblst = 
	  List.map (function x -> {id= List.assoc x.id new_vpair;
				   body= alpha_conv (new_vpair@vpair) x.body;})
	    blst
	in 
	Letrec(newblst, alpha_conv (new_vpair@vpair) exp)
    | If(e1,e2,e3) ->
	If(alpha_conv vpair e1,
	   alpha_conv vpair e2,
	   alpha_conv vpair e3)
    | Set _ -> failwith "invalid syntax."
    (*| _ -> failwith "not yet implemented"*)
	  
and alpha_conv_sval vpair sval =
  match sval with
    Number(n) -> sval
  | Boolean(b) -> sval
  | Character(c) -> sval
  | String(str) -> sval
  | Symbol(s) -> sval
  | Null -> sval
  | Undef -> sval

let string_of_svalue sval =
  match sval with
    Number(n) -> string_of_scmnum n
  | Boolean(b) -> if b then "#t" else "#f"
  | Character(c) -> let s = "a" in s.[0] <- c ; "#\\" ^ s
  | String(str) -> "\"" ^ str ^ "\""
  | Symbol(s) -> s
  | Null -> "'()"
  | Undef -> "*undef*"
  
(* ----------- ɽ --------------*)   
open Format
let print_exp (e:exp) =
  let rec pr_texp ppf texp = 
    match texp with 
      Val(v) -> fprintf ppf "%s" (string_of_svalue v)
    | Var(v) ->  fprintf ppf "%s" v 
    | Lambda((args,Some tail),body) ->
	let sargs = (List.fold_right (fun x y -> x^" "^y)  args "")
	    ^ (" . " ^ tail)
	in
	fprintf ppf "@[<hov 3>(lambda@ %s@ %a)@]" 
	  ("("^sargs^")") pr_texp body 
    | Lambda((args,None),body) ->
	let sargs = (List.fold_right (fun x y -> x^" "^y)  args "")
	in
	fprintf ppf "@[<hov 3>(lambda@ %s@ %a)@]" 
	  ("("^sargs^")") pr_texp body 
    | App(f,args) ->
	fprintf ppf "(@[<hov 0>%a@ %a@])" 
	  pr_texp f pr_lst args
    | Let(tbindlst,body) ->
	fprintf ppf "@[<hov 3>(let@ (@[<h 0>%a)@])%a@]" 
	  pr_tblst tbindlst pr_texp body
    | Letrec(bindlst,body) ->
	fprintf ppf "@[<hov 3>(letrec@ (@[<h 0>%a)@]%a)@]" 
	  pr_tblst bindlst pr_texp body 
    | If(e1,e2,e3) ->
	fprintf ppf "@[<hov 0>(if@ %a@ %a@ %a)@]" 
	  pr_texp e1 pr_texp e2 pr_texp e3 
    | Set(v,e)  ->
	fprintf ppf "@[<hov 0>(set!@ %s@ %a)@]" 
	  v pr_texp e
	  
(*    | TPair(e1,e2,st) ->
	fprintf ppf "@[<hov 0>(%a@ .@ %a)@,:%s@]" 
	  pr_texp e1 pr_texp e2 (string_of_stype st) *)
  and pr_lst ppf =
    function 
	hd::tl -> fprintf ppf  "@[<hov 0>%a@,@ %a@]"  pr_texp hd pr_lst tl
      | [] -> fprintf ppf "%s" ""
  and pr_tbind ppf b = 
    fprintf ppf "@[<hov 1>(%s@,%a)@]" 
      b.id  pr_texp b.body 
  and pr_tblst ppf =
    function 
	hd::tl -> fprintf ppf "@[<hov 0>%a@,%a@]"  pr_tbind hd pr_tblst tl
      | [] -> fprintf ppf "%s" ""
  in 
  printf "%a@." pr_texp e; print_string "\n"
    
  

(* --- remove set! ---- *)

(* set!оΤȤʤ뤳Ȥ狼äѿˤĤƽ񤭴 *)
let rec replace_set_and_setcar varname exp =
  (* print_endline ("replace_setvar" ^ varname);  *)
  match exp with
    Val(_) -> exp 
  | Var(x) when x = varname -> App(Var("car"),[Var(x)])
  | Var(x) -> Var(x)
  | Lambda((args,Some(x)),body) ->
      if List.mem varname (x::args) 
      then exp
      else Lambda((args,Some(x)),replace_set_and_setcar varname body)
  | Lambda((args,None),body) ->
      if List.mem varname args
      then exp
      else Lambda((args,None),replace_set_and_setcar varname body)
  | App(f,args) -> App(replace_set_and_setcar varname f,
		       List.map (replace_set_and_setcar varname) args)
  | Let(binds,body) -> 
      let bind_vars = List.map (function x -> x.id) binds in
      Let(List.map (function x -> {id=x.id; body=replace_set_and_setcar varname x.body}) binds,
	  if List.mem varname bind_vars
	  then body 
	  else replace_set_and_setcar varname body)
  | Letrec(binds,body) -> 
      let bind_vars = List.map (function x -> x.id) binds in
      if List.mem varname bind_vars 
      then 
	Letrec(binds,body)
      else 
	Letrec(List.map (function x -> {id=x.id; body=replace_set_and_setcar varname x.body}) binds,
	       replace_set_and_setcar varname body)
  | If(e1,e2,e3) -> If(replace_set_and_setcar varname e1,
		       replace_set_and_setcar varname e2,
		       replace_set_and_setcar varname e3)
  | Set(x,body) when x = varname -> App(Var("set-car!"),
					 [Var(x);replace_set_and_setcar varname body])
  | Set(x,body) -> Set(x,replace_set_and_setcar varname body)
  
(* set!оΤȤʤѿ򸫤Ĥ *)
let rec find_setvar v body =
  (* print_endline ("find_setvar " ^ v); 
  print_exp body;*)
  match body with 
    Val(_) -> false
  | Var(_) -> false
  | Lambda((args,Some(x)),body) ->
      if List.mem v (x::args)
      then false
      else find_setvar v body 
  | Lambda((args,None),body) ->
      if List.mem v args
      then false
      else find_setvar v body 
  | App(f,args) ->
      List.fold_right (||) (List.map (find_setvar v) (f::args)) false 
  | Let(binds,exp) ->
      let flag = List.fold_right (||) (List.map (function x -> find_setvar v x.body) binds) false  in
      if flag 
      then true
      else 
	let bind_vars = List.map (function x -> x.id) binds in
	if List.mem v bind_vars
	then 
	  false
	else 
	  find_setvar v exp
  | Letrec(binds,exp) ->
      let bind_vars = List.map (function x -> x.id) binds in
      if List.mem v bind_vars
      then 
	false 
      else 
	((List.fold_right (||) (List.map (function x -> find_setvar v x.body) binds) false)
       ||
	 find_setvar v exp)
  | If(e1,e2,e3) -> (find_setvar v e1)||(find_setvar v e2)||(find_setvar v e3)
  | Set(sv,exp) ->
      if v = sv 
      then true
      else find_setvar v exp
	

let rec remove_set exp = 
  (* print_endline "remove_set";  *)
  match exp with
    Val(_) -> exp 
  | Var(_) -> exp 
  | Lambda((args,Some(x)),body) ->
      Lambda((args,Some(x)),
	     remove_set (List.fold_right 
			   (fun x b -> 
			     if (find_setvar x b)
			     then Let([{id=x;body=App(Var("cons"),[Var(x);Val(Null)])}],
				      replace_set_and_setcar x b)
			     else b) (x::args) body))
  | Lambda((args,None),body) ->
      Lambda((args,None),
	     remove_set (List.fold_right 
			   (fun x b -> 
			     if (find_setvar x b)
			     then 
			       begin
				 print_endline ("remove " ^ x);
				 Let([{id=x;body=App(Var("cons"),[Var(x);Val(Null)])}],
				     replace_set_and_setcar x b)
			       end
			     else b) args body))
  | App(f,args) -> App(remove_set f,List.map remove_set args)
  | Let(binds,exp) ->
      let bind_vars = List.map (function x -> x.id) binds in
      let set_vars = List.filter (function v -> find_setvar v exp) bind_vars in
      Let(
      List.map (function b -> 
	if List.mem b.id set_vars 
	then {id=b.id;body=App(Var("cons"),[remove_set b.body;Val(Null)])}
	else {id=b.id;body=remove_set b.body})
	binds,
      remove_set (List.fold_right (fun v body -> replace_set_and_setcar v body) set_vars exp))
  | Letrec(binds,exp) ->
      let bind_vars = List.map (function x -> x.id) binds in
      let bind_bodys = List.map (function x -> x.body) binds in
      
      let set_vars = 
	List.filter 
	  (function v -> List.fold_right (||) (List.map (find_setvar v) (exp::bind_bodys)) false)
	  bind_vars in
      Letrec(
      List.map (function b -> 
	let body = 
	  remove_set (List.fold_right 
			(fun v body -> replace_set_and_setcar v body) set_vars b.body) in
	if List.mem b.id set_vars 
	then {id=b.id;body=App(Var("cons"),[body;Val(Null)])}
	else {id=b.id;body=body})
	binds,
      remove_set (List.fold_right (fun v body -> replace_set_and_setcar v body) set_vars exp))
  | If(e1,e2,e3) -> If(remove_set e1, remove_set e2, remove_set e3)
  | Set _ -> failwith "Invalid syntax.(syntax.ml : remove_set)"

