open Syntax
open Stype
  

(* $B7?$D$-%7%s%?%C%/%9(B *) 
type texp =
    TVal of svalue * stype
  | TVar of string * stype
  | TLambda of (lbind list * lbind option) * texp * stype
  | TApp of texp * texp list * stype
  | TLet of tbind list * texp * stype
  | TLetrec of tbind list * texp * stype
  | TIf  of texp * texp * texp * stype 
and tbind = { sid:string ; st:stype_scheme ;sbody:texp }
and lbind = { lid:string ; lt:stype_scheme}

let rec size t =
  match t with
    TVal(_,_) -> 1
  | TVar(_,_) -> 1
  | TLambda(_,t,_) -> (size t) +1
  | TApp(f,args,_) -> List.fold_right (+) (List.map size (f::args)) 1
  | TLet(bnds,e,s) -> 
      List.fold_right (+) (List.map (function x -> size x.sbody) bnds) (size e)
  | TLetrec(bnds,e,s) -> 
	List.fold_right (+) (List.map (function x -> size x.sbody) bnds) (size e)
  | TIf(a,b,c,s) -> (size a) + (size b) + (size c) + 1
  

let getType (te:texp) :stype = 
  match te with
    TVal(_,s) -> s
  | TVar(_,s) -> s
  | TLambda(_,_,s) -> s
  | TApp(_,_,s) -> s
  | TLet(_,_,s) -> s
  | TLetrec(_,_,s) -> s
  | TIf(_,_,_,s) -> s

let isValue (te:texp) : bool = 
  match te with
    TVal(_,_) -> false
  | TVar(_,_) -> false
  | TLambda(_,_,_) -> true
  | TApp(_,_,_) -> false
  | TLet(_,_,_) -> false
  | TLetrec(_,_,_) -> false
  | TIf(_,_,_,_) -> false

let rec substitute_texp (sb:subst) (te:texp) :texp = 
  match te with
    TVal(a,s) -> TVal(a,substitute sb s) 
  | TVar(a,s) -> TVar(a,substitute sb s) 
  | TLambda((arg,tail),b,s) -> 
      TLambda((List.map 
		(function x -> 
		  {lid=x.lid; lt=substitute_scheme sb x.lt}) arg,
	       match tail with
		 Some t ->
		   Some{lid=t.lid; lt=substitute_scheme sb t.lt}
	       | None -> None),
	       substitute_texp sb b,substitute sb s)
  | TApp(a,b,s) -> TApp(substitute_texp sb a,
			List.map (substitute_texp sb) b,
			substitute sb s) 
  | TLet(a,b,s) -> 
      TLet(List.map 
	     (function x-> 
	       {sid=x.sid;   st=substitute_scheme sb x.st;
		sbody=substitute_texp sb x.sbody}) a,
	   substitute_texp sb b,
	   substitute sb s) 
  | TLetrec(a,b,s) -> 
      TLetrec(List.map 
	     (function x-> 
	       {sid=x.sid;   st=substitute_scheme sb x.st;
		sbody=substitute_texp sb x.sbody}) a,
	   substitute_texp sb b,
	   substitute sb s) 
  | TIf(a,b,c,s) -> TIf(substitute_texp sb a,
			substitute_texp sb b,
			substitute_texp sb c,
			substitute sb s) 


let rec type_of_svalue (s:svalue) =
  match s with
    Number(n) -> num_type ()
  | Boolean(b) -> bool_type ()
  | Character(c) -> char_type ()
  | String(str) -> string_type ()
  | Symbol(s) -> symbol_type ()
  | Null -> null_type ()
  | Undef -> unknown_type ()
	

let rec stype_infer (env:type_env) (exp:exp) : (subst * texp) =
  match exp with 
    Val(v) ->
      ([],TVal(v,type_of_svalue v))
  | Var(s) ->
      begin 
	try
	  let scm = List.assoc s env in
	  let t = instantiate scm in
	  ([],TVar(s,t))
	with Not_found -> failwith "TSyntax.stype_infer : unbound variable"
      end
  | Lambda((arglst,None),exp) ->
      let newarg = 
	List.map 
	  (function x -> {lid=x;lt=stype2scheme (unknown_type ())})
	  arglst in
      let newenv = List.map (function x-> (x.lid,x.lt)) newarg in
      let (s,rettexp) = stype_infer (newenv@env) exp in
      let rettype = getType rettexp in
      let newarg2 = 
	List.map (function x -> {lid=x.lid;lt=substitute_scheme s x.lt}) 
	  newarg in
      let tailtype = list_type (unknown_type ()) in
      let lamtype = 
	func_type ((List.map (function x-> instantiate x.lt) newarg2),tailtype) rettype in
      (s,TLambda((newarg2,None),rettexp,lamtype))
  | Lambda((arglst,Some tail),exp) ->
      let newarg = 
	List.map 
	  (function x -> {lid=x;lt=stype2scheme (unknown_type ())})
	  arglst in
      let newtail = {lid=tail; lt= stype2scheme (list_type (unknown_type ()))} in
      let newenv = (newtail.lid,newtail.lt)::(List.map (function x-> (x.lid,x.lt)) newarg) in
      let (s,rettexp) = stype_infer (newenv@env) exp in
      let rettype = getType rettexp in
      let newarg2 = 
	List.map (function x -> {lid=x.lid;lt=substitute_scheme s x.lt}) 
	  newarg in
      let newtail2 = {lid=newtail.lid;lt=substitute_scheme s newtail.lt} in
      let lamtype = 
	func_type ((List.map (function x-> instantiate x.lt) newarg2),
		   instantiate newtail2.lt) rettype in
      (s,TLambda((newarg2,Some newtail2),rettexp,lamtype))
  | App(f,args) ->
      let (sf,tfe) = stype_infer env f in
      let rec iter (s:subst) args :(subst * texp list)=
	match args with
	  hd::tl ->
	    let (newsub,ts) = stype_infer (substitute_tenv s env) hd in
	    let (tails,tailts) = iter (subst_combine newsub s) tl  in
	    (tails, ts::tailts)
	| [] -> (s,[])
      in
      let (sarg,targs) = iter sf args in
      let rettype = unknown_type () in
      let tailtype = list_type (unknown_type ()) in
      let t1 = func_type ((List.map getType targs),tailtype) rettype in
      let sfinal = unify (getType (substitute_texp sarg tfe)) t1 in
(*      print_endline "unify";
      print_endline (string_of_stype (getType (substitute_texp sarg tfe)));
      print_endline (string_of_stype t1);
      print_string "subst ";print_substs sfinal;
      print_endline "after"; 
      print_endline (string_of_stype (substitute sfinal t1)); *)

      let sfinal' = subst_combine sfinal sarg in
      (sfinal',TApp(substitute_texp sfinal' tfe, 
		   List.map (substitute_texp sfinal') targs,
		   substitute sfinal' rettype))
  | Let(blst,body) ->
      let rec iter (sb:subst) (lst:bind list): 
	  (subst * (string * texp) list) =
	match lst with
	  hd::tl ->
	    let (newsb,newt) = stype_infer
		(substitute_tenv sb env) hd.body in
	    let (isb, tblst) =
	      iter (subst_combine newsb sb) tl in
	    (isb,(hd.id,substitute_texp isb newt)::tblst)
	| [] -> (sb,[])
      in
      let (newsb,tblst) = iter [] blst in
      let baseenv = substitute_tenv newsb env in
      let tblst =
	List.map
	  (function (id,tep) ->
	    match tep with
	      TLambda(_,_,_) -> (* lambda $B$@$C$?$i(Bgen *)
		{sid=id;
		 st=
		 begin
		   let ret = generalize baseenv (substitute newsb (getType tep)) in
		   (* print_endline (string_of_stype_scheme ret); *)
		   ret
		 end;
		 sbody=tep}
	    | _ ->
		{sid=id;
		 st=stype2scheme (substitute newsb (getType tep));
		 sbody=tep})
	  tblst
      in
      let newenv = List.map (function x -> (x.sid,x.st)) tblst in
      let (sbody,tbody) = stype_infer (newenv@baseenv) body in
      (subst_combine sbody newsb,
       TLet(List.map 
	      (function x -> {sid =  x.sid;
			      st =   substitute_scheme sbody x.st;
			      sbody =substitute_texp sbody x.sbody
			    })
	      tblst,
	    tbody,getType tbody))
  | Letrec(blst,body) ->
      let (newbenv:(string*stype_scheme) list) = List.map 
	  (function b -> (b.id, stype2scheme(unknown_type ()))) blst in
      (* $BB+G{ItJ,$N7??dO@(B *)
      let rec iter (sb:subst) (lst:bind list): (subst * (string * texp) list) =
	match lst with
	  hd::tl -> 
	    let (newsb,newt) =
	      (* body $BItJ,$r?dO@(B *)
	      stype_infer (substitute_tenv sb (newbenv@env)) hd.body in
	    let newsb2 = 
	      subst_combine 
		(unify 
		   (instantiate (substitute_scheme newsb 
				   (List.assoc (hd.id:string) newbenv )) )
		   (getType newt))
		newsb
	    in
	    let (isb, tblst) = 
	      iter (subst_combine newsb2 sb) tl in
	    (isb,(hd.id, substitute_texp isb newt)::tblst) 
	| [] -> (sb,[])
      in 
      let (newsb,tblst) = iter [] blst in (*$BB+G{$N7??dO@(B*)
      (* $B4D6-$r?7$7$/$9$k(B *)
      let baseenv = substitute_tenv newsb env in
      (* print_endline "free vars";
      List.iter (function (x,st) -> print_endline (x ^ " : " ^ (string_of_stype_scheme st))) baseenv;
      print_endline "";*)
      let tmpenv = List.map (function (id,tep) -> (id,stype2scheme (getType tep))) tblst in
      (* generalize *)
      let newbind = 
	List.map 
	  (function (id,tep) -> 
	    match tep with
	      TLambda(_,_,_) ->
		{sid=id; 
		 st=
		 begin 
		   let a = generalize (tmpenv@baseenv) (getType tep) in
		   (* print_endline (string_of_stype_scheme a); *)
		   a
		 end;
		 sbody=tep}
	    | _ -> {sid=id; st=stype2scheme (getType tep); sbody=tep})
	  tblst in
      (* body $B$r?dO@(B *)
      let newenv = List.map (function x -> (x.sid,x.st)) newbind in
      let (sbody,tbody) = stype_infer (newenv@baseenv) body in
      (subst_combine sbody newsb,
       TLetrec(List.map 
		 (function x -> {sid=x.sid; 
				 st=substitute_scheme sbody x.st;
				 sbody=substitute_texp sbody x.sbody
			       })
		 newbind,
	       tbody,getType tbody))
  | If(e1,e2,e3) ->
      let (sb1,t1) = stype_infer env e1 in
      let (sb2,t2) = stype_infer (substitute_tenv sb1 env) e2 in
      let (sb3,t3) = stype_infer 
	  (substitute_tenv (subst_combine sb1 sb2) env) e3 in
      let sb4 = unify (substitute sb3 (getType t2)) (getType t3) in
      let sb5 = (subst_combine sb4
		   (subst_combine sb3
		      (subst_combine sb2 sb1))) in
      (sb5,TIf(substitute_texp sb5 t1,
	       substitute_texp sb5 t2,
	       substitute_texp sb5 t3,
	       substitute sb5 (getType t2) ))
  | Set _ -> failwith "Invalid syntax"
(*  | Pair(e1,e2) ->
      let (sb1,t1) = stype_infer env e1 in
      let (sb2,t2) = stype_infer (substitute_tenv sb1 env) e2 in
      let sb3 = subst_combine sb1 sb2 in
      (sb3,
       TPair(substitute_texp sb3 t1,
	     substitute_texp sb3 t2,
	     pair_type (substitute sb3 (getType t1)) 
	       (substitute sb3 (getType t2)))) *)
(*  | _ -> failwith "now printing" *)
 


(* $BI=<(4X?t(B *)
let rec string_of_texp =
  function
      TVal(v,st) -> (string_of_svalue v) ^ ":" ^ (string_of_stype st)
    | TVar(v,st) -> v ^ ":" ^ (string_of_stype st)
    | TLambda((lblst,Some tail),te,st) ->
	"(lambda (" ^ (List.fold_right (^) (List.map string_of_lbind lblst) "")
	^ " . " ^ (string_of_lbind tail)
	^ ")" ^(string_of_texp te) ^ "):" ^ (string_of_stype st)
    | TLambda((lblst,None),te,st) ->
	"(lambda (" ^ (List.fold_right (^) (List.map string_of_lbind lblst) "")
	^ ")" ^(string_of_texp te) ^ "):" ^ (string_of_stype st)
    | TApp(tf,telst,st) ->
	(string_of_texp tf) 
	^ (List.fold_right (^) (List.map string_of_texp telst) "")
	^ ":" ^ (string_of_stype st)
    | TLet(tblst,te,st) ->
	"(let (" ^ (List.fold_right (^) (List.map string_of_tbind tblst) "")
	^ ")" ^ (string_of_texp te) ^ "):" ^ (string_of_stype st)
    | TLetrec(tblst,te,st) ->
	"(letrec (" ^ (List.fold_right (^) (List.map string_of_tbind tblst) "")
	^ ")" ^ (string_of_texp te) ^ "):" ^ (string_of_stype st)
    | TIf(te1,te2,te3,st) ->
	"(if " ^ (List.fold_right (^) (List.map string_of_texp [te1;te2;te3]) "")
	^ "):" ^ (string_of_stype st)
(*    | TPair(te1,te2,st) ->
	"(" ^ (string_of_texp te1) ^ " . " ^ (string_of_texp te2)
	^ "):" ^ (string_of_stype st) *)
and string_of_lbind lb =
  lb.lid 
and string_of_tbind tb =
  "(" ^ tb.sid ^ ":" ^ (string_of_stype_scheme tb.st) ^ " " ^ (string_of_texp tb.sbody) ^ ")"
											    
(* ------------------------------------------------- *)  
open Format
let print_texp (e:texp) =
  let rec pr_texp ppf texp = 
    match texp with 
      TVal(v,st) -> fprintf ppf "%s:%s" (string_of_svalue v) (string_of_stype st)
    | TVar(v,st) ->  fprintf ppf "%s:%s" v (string_of_stype st)
    | TLambda((args,Some tail),body,st) ->
	let sargs = (List.fold_right (fun x y -> x.lid^" "^y)  args "")
	    ^ (" . " ^ tail.lid)
	in
	fprintf ppf "@[<hov 3>(lambda@ %s@ %a)@,:%s@]" 
	  ("("^sargs^")") pr_texp body (string_of_stype st)
    | TLambda((args,None),body,st) ->
	let sargs = (List.fold_right (fun x y -> x.lid^" "^y)  args "")
	in
	fprintf ppf "@[<hov 3>(lambda@ %s@ %a)@,:%s@]" 
	  ("("^sargs^")") pr_texp body (string_of_stype st)
    | TApp(f,args,st) ->
	fprintf ppf "(@[<hov 0>%a@ %a@])@,:%s" 
	  pr_texp f pr_lst args (string_of_stype st)
    | TLet(tbindlst,body,st) ->
	fprintf ppf "@[<hov 3>(let@ (@[<h 0>%a)@]@,%a)@,:%s@]" 
	  pr_tblst tbindlst pr_texp body (string_of_stype st)
    | TLetrec(bindlst,body,st) ->
	fprintf ppf "@[<hov 3>(letrec@ (@[<h 0>%a)@]@,%a)@,:%s@]" 
	  pr_tblst bindlst pr_texp body (string_of_stype st)
    | TIf(e1,e2,e3,st) ->
	fprintf ppf "@[<hov 0>(if@ %a@ %a@ %a)@,:%s@]" 
	  pr_texp e1 pr_texp e2 pr_texp e3 (string_of_stype st)
(*    | TPair(e1,e2,st) ->
	fprintf ppf "@[<hov 0>(%a@ .@ %a)@,:%s@]" 
	  pr_texp e1 pr_texp e2 (string_of_stype st) *)
  and pr_lst ppf =
    function 
	hd::tl -> fprintf ppf  "@[<hov 0>%a@,@ %a@]"  pr_texp hd pr_lst tl
      | [] -> fprintf ppf "%s" ""
  and pr_tbind ppf b = 
    fprintf ppf "@[<hov 1>(%s:%s@,%a)@]" 
      b.sid (string_of_stype_scheme b.st) pr_texp b.sbody 
  and pr_tblst ppf =
    function 
	hd::tl -> fprintf ppf "@[<hov 0>%a@,%a@]"  pr_tbind hd pr_tblst tl
      | [] -> fprintf ppf "%s" ""
  in 
  printf "%a@." pr_texp e; print_string "\n"
    
  
